{-# LANGUAGE TypeOperators #-}
{-
ToDo:
Do our implementations prevent recomputing of shape sizes, via shapeOffset?
-}
{-
ToDo:
Any relation to Shape.DataPattern?
-}
{-
ToDo:
expose Cons only via an Unchecked module
this could also provide unchecked 'mapShape'
-}
module Numeric.BLAS.Subobject.View where

import qualified Numeric.BLAS.Subobject.Layout as Layout
import Numeric.BLAS.Subobject.Layout ((*.))

import qualified Data.Array.Comfort.Shape as Shape
import qualified Data.Array.Comfort.Boxed as BoxedArray
import Data.Array.Comfort.Shape ((::+)((::+)))

import qualified Data.Traversable as Trav
import qualified Data.List as List
import qualified Data.Ix as Ix
import Data.IntMap (IntMap)
import Data.Map (Map)
import Data.Ix (Ix)
import Data.Bool.HT (if')

import Text.Printf (printf)


{- $setup
>>> :set -XTypeOperators
>>>
>>> import qualified Numeric.BLAS.Subobject.View as Slice
>>> import qualified Numeric.BLAS.Subobject.Layout as Layout
>>> import Test.Slice (shapeInt)
>>>
>>> import qualified Data.Array.Comfort.Boxed as Array
>>> import qualified Data.Array.Comfort.Shape as Shape
>>> import qualified Data.Map as Map
>>> import Data.Array.Comfort.Shape ((::+)((::+)))
>>> import Data.Array.Comfort.Boxed ((!))
>>>
>>> import Control.Applicative (liftA2, liftA3, pure, (<$>))
>>>
>>> import qualified Test.QuickCheck as QC
>>>
>>> genShape :: QC.Gen (Shape.Range Int)
>>> genShape =
>>>    liftA2
>>>       (\m n -> Shape.Range (min m n) (max m n))
>>>       QC.arbitrary QC.arbitrary
>>>
>>> genAppend :: QC.Gen (Shape.Range Int ::+ Shape.Range Int)
>>> genAppend = liftA2 (::+) genShape genShape
>>>
>>> type Slice = Slice.T Layout.Slice
>>>
>>> genSlice :: sh -> QC.Gen (Slice sh)
>>> genSlice sh =
>>>    liftA3 Slice.Cons
>>>       (QC.choose (0,100))
>>>       (Layout.Slice <$> QC.choose (1,100))
>>>       (pure sh)
>>>
>>> genSlice2 :: shA -> shB -> QC.Gen (Slice shA, Slice shB)
>>> genSlice2 shA shB = do
>>>    s <- QC.choose (0,100)
>>>    k <- Layout.Slice <$> QC.choose (1,100)
>>>    return (Slice.Cons s k shA, Slice.Cons s k shB)
>>>
>>> type ShapeInt = Shape.ZeroBased Int
>>>
>>> toShapeInt :: (Shape.C sh) => Slice sh -> Slice ShapeInt
>>> toShapeInt = Slice.mapShape (shapeInt . Shape.size)
-}


data T lay sh = Cons {start :: Int, layout :: lay, shape :: sh}
   deriving (Eq, Show)

type Subvector = T Layout.Subvector
-- type Slice = T Layout.Slice
type Submatrix = T Layout.Submatrix


class C lay where
   rowOffset :: (Shape.C sh) => lay -> sh -> Int
   elemInc :: lay -> Int
   scaleLeadingDim :: Int -> lay -> lay

instance C Layout.Subvector where
   rowOffset _ = Shape.size
   elemInc Layout.Subvector = 1
   scaleLeadingDim _ = id

instance C Layout.Slice where
   rowOffset (Layout.Slice k) sh = k * Shape.size sh
   elemInc (Layout.Slice k) = k
   scaleLeadingDim _ = id

{-
Attention:
You must only use this on @(height,width)@ or @Shape.Square sh@ shaped arrays,
and the leading dimension must refer to the width of that shape.
-}
instance C Layout.Submatrix where
   rowOffset (Layout.Submatrix ld) _sh = ld
   elemInc (Layout.Submatrix _) = 1
   scaleLeadingDim j (Layout.Submatrix ld) = Layout.Submatrix $ j*ld


class (C lay) => Slice lay where
   defltLayout :: lay
   sliceFromLayout :: lay -> Layout.Slice

instance Slice Layout.Subvector where
   defltLayout = Layout.Subvector
   sliceFromLayout Layout.Subvector = Layout.Slice 1

instance Slice Layout.Slice where
   defltLayout = Layout.Slice 1
   sliceFromLayout = id



fromShape :: (Slice lay, Shape.C sh) => sh -> T lay sh
fromShape = Cons 0 defltLayout

toSlice :: (Slice lay) => T lay sh -> T Layout.Slice sh
toSlice (Cons s k sh) = Cons s (sliceFromLayout k) sh


reshape :: (Shape.C sh0, Shape.C sh1) => sh1 -> T lay sh0 -> T lay sh1
reshape sh1 (Cons s k sh0) =
   Cons s k $
   if Shape.size sh0 == Shape.size sh1
      then sh1
      else error "View.reshape: shape sizes mismatch"

mapShape :: (Shape.C sh0, Shape.C sh1) => (sh0 -> sh1) -> T lay sh0 -> T lay sh1
mapShape f slc = reshape (f $ shape slc) slc


{- ToDo: make a special version for Submatrix
row ::
   (Shape.Indexed sh0, Shape.C sh1) =>
   Shape.Index sh0 -> T Submatrix (sh0,sh1) -> T Subvector sh1
-}
row ::
   (Slice lay, Shape.Indexed sh0, Shape.C sh1) =>
   Shape.Index sh0 -> T lay (sh0,sh1) -> T lay sh1
row ix0 (Cons s k (sh0,sh1)) =
   Cons (s + Shape.offset sh0 ix0 * rowOffset k sh1) k sh1

column ::
   (Shape.C sh0, Shape.Indexed sh1) =>
   Shape.Index sh1 -> T Layout.Slice (sh0,sh1) -> T Layout.Slice sh0
column ix1 (Cons s k (sh0,sh1)) =
   let (size1, offset1) = Shape.sizeOffset sh1
   in Cons (s + elemInc k * offset1 ix1) (k *. size1) sh0

{- |
prop> :{
   QC.forAll (QC.choose (1,100)) $ \numRows ->
   QC.forAll (QC.choose (0,100)) $ \numColumns ->
   QC.forAll (genSlice (shapeInt numRows, shapeInt numColumns)) $ \slice ->
   QC.forAll (QC.elements $ Shape.indices $ shapeInt numRows) $ \ix ->
   Slice.row ix slice == Slice.rowArray slice ! ix
:}
-}
rowArray ::
   (Slice lay, Shape.C sh0, Shape.C sh1) =>
   T lay (sh0,sh1) -> BoxedArray.Array sh0 (T lay sh1)
rowArray slc =
   BoxedArray.fromList (fst $ shape slc) $ rows slc

rows ::
   (Slice lay, Shape.C sh0, Shape.C sh1) =>
   T lay (sh0,sh1) -> [T lay sh1]
rows (Cons s k (sh0,sh1)) =
   let step = rowOffset k sh1
   in List.map (\si -> Cons si k sh1) $
      List.take (Shape.size sh0) $ iterate (step+) s

{- |
prop> :{
   QC.forAll (QC.choose (0,100)) $ \numRows ->
   QC.forAll (QC.choose (1,100)) $ \numColumns ->
   QC.forAll (genSlice (shapeInt numRows, shapeInt numColumns)) $ \slice ->
   QC.forAll (QC.elements $ Shape.indices $ shapeInt numColumns) $ \ix ->
   Slice.column ix slice == Slice.columnArray slice ! ix
:}
-}
columnArray ::
   (Shape.C sh0, Shape.C sh1) =>
   T Layout.Slice (sh0,sh1) -> BoxedArray.Array sh1 (T Layout.Slice sh0)
columnArray slc =
   BoxedArray.fromList (snd $ shape slc) $ columns slc

columns ::
   (Shape.C sh0, Shape.C sh1) =>
   T Layout.Slice (sh0,sh1) -> [T Layout.Slice sh0]
columns (Cons s k (sh0,sh1)) =
   let step = k *. Shape.size sh1
   in List.map (\si -> Cons si step sh0) $
      List.take (Shape.size sh1) $ iterate (elemInc k +) s


{- |
prop> :{
   QC.forAll (QC.choose (1,100)) $ \numTopRows ->
   QC.forAll (QC.choose (1,100)) $ \numBottomRows ->
   QC.forAll (QC.choose (0,100)) $ \numColumns ->
   QC.forAll (genSlice (shapeInt numTopRows ::+ shapeInt numBottomRows,
                        shapeInt numColumns)) $ \slice ->
   Slice.rowArray slice
   ==
   Array.append
      (Slice.rowArray (Slice.topSubmatrix slice))
      (Slice.rowArray (Slice.bottomSubmatrix slice))
:}
-}
topSubmatrix ::
   (C lay, Shape.C sh, Shape.C sh0, Shape.C sh1) =>
   T lay (sh0::+sh1, sh) -> T lay (sh0,sh)
topSubmatrix (Cons s k (sh0::+_sh1, sh)) =
   Cons s k (sh0,sh)

bottomSubmatrix ::
   (C lay, Shape.C sh, Shape.C sh0, Shape.C sh1) =>
   T lay (sh0::+sh1, sh) -> T lay (sh1,sh)
bottomSubmatrix (Cons s k (sh0::+sh1, sh)) =
   Cons (s + Shape.size sh0 * rowOffset k sh) k (sh1,sh)


diagonal :: (C lay, Shape.C sh) =>
   T lay (Shape.Square sh) -> T Layout.Slice sh
diagonal (Cons s k (Shape.Square sh)) =
   Cons s (Layout.Slice $ rowOffset k sh + elemInc k) sh

cartesianFromSquare :: T lay (Shape.Square sh) -> T lay (sh,sh)
cartesianFromSquare (Cons s k (Shape.Square sh)) = Cons s k (sh,sh)

squareRow ::
   (Slice lay, Shape.Indexed sh) =>
   Shape.Index sh -> T lay (Shape.Square sh) -> T lay sh
squareRow ix0 = row ix0 . cartesianFromSquare

squareColumn ::
   (Shape.Indexed sh) =>
   Shape.Index sh -> T Layout.Slice (Shape.Square sh) -> T Layout.Slice sh
squareColumn ix1 = column ix1 . cartesianFromSquare



assocOuter ::
   (Slice lay, Shape.Indexed sh0, Shape.C sh1, Shape.C sh2) =>
   T lay (sh0,(sh1,sh2)) -> T lay ((sh0,sh1),sh2)
assocOuter (Cons s k (sh0,(sh1,sh2))) = Cons s k ((sh0,sh1),sh2)

assocInner ::
   (C lay, Shape.Indexed sh0, Shape.C sh1, Shape.C sh2) =>
   T lay ((sh0,sh1),sh2) -> T lay (sh0,(sh1,sh2))
assocInner (Cons s k ((sh0,sh1),sh2)) =
   Cons s (scaleLeadingDim (Shape.size sh1) k) (sh0,(sh1,sh2))


plane12 ::
   (Slice lay, Shape.Indexed sh0, Shape.C sh1, Shape.C sh2) =>
   Shape.Index sh0 -> T lay (sh0,sh1,sh2) -> T lay (sh1,sh2)
plane12 ix0 (Cons s k (sh0,sh1,sh2)) =
   Cons (s + elemInc k * Shape.offset sh0 ix0 * Shape.size (sh1,sh2)) k (sh1,sh2)

plane01 ::
   (Slice lay, Shape.C sh0, Shape.C sh1, Shape.Indexed sh2) =>
   Shape.Index sh2 -> T lay (sh0,sh1,sh2) -> T Layout.Slice (sh0,sh1)
plane01 ix2 (Cons s k (sh0,sh1,sh2)) =
   let (size2, offset2) = Shape.sizeOffset sh2
   in Cons (s + elemInc k * offset2 ix2) (sliceFromLayout k *. size2) (sh0,sh1)

column2of3 ::
   (Slice lay, Shape.Indexed sh0, Shape.Indexed sh1, Shape.C sh2) =>
   Shape.Index sh0 -> Shape.Index sh1 -> T lay (sh0,sh1,sh2) -> T lay sh2
column2of3 ix0 ix1 = row ix1 . plane12 ix0

column1of3 ::
   (Slice lay, Shape.Indexed sh0, Shape.C sh1, Shape.Indexed sh2) =>
   Shape.Index sh0 -> Shape.Index sh2 ->
   T lay (sh0,sh1,sh2) -> T Layout.Slice sh1
column1of3 ix0 ix2 = column ix2 . toSlice . plane12 ix0

column0of3 ::
   (Slice lay, Shape.C sh0, Shape.Indexed sh1, Shape.Indexed sh2) =>
   Shape.Index sh1 -> Shape.Index sh2 ->
   T lay (sh0,sh1,sh2) -> T Layout.Slice sh0
column0of3 ix1 ix2 = column ix1 . plane01 ix2
{-
ToDo:
Shape.Cube
required: Shape.cubeToCartesian
-}

left :: (Slice lay, Shape.C sh0, Shape.C sh1) => T lay (sh0::+sh1) -> T lay sh0
left (Cons s k (sh0::+_sh1)) = Cons s k sh0

right :: (Slice lay, Shape.C sh0, Shape.C sh1) => T lay (sh0::+sh1) -> T lay sh1
right (Cons s k (sh0::+sh1)) = Cons (s + elemInc k * Shape.size sh0) k sh1


type ShapeInt = Shape.ZeroBased Int

{- |
prop> :{
   QC.forAll (genSlice =<< genAppend) $ \slice ->
   Slice.take
      (case Slice.shape slice of (sh::+_) -> Shape.size sh)
      (toShapeInt slice)
   ==
   toShapeInt (Slice.left slice)
:}

It is an unchecked error,
when the values of type @n@ exceed the range of @Int@.
-}
take :: (Slice lay, Integral n) =>
   n -> T lay (Shape.ZeroBased n) -> T lay (Shape.ZeroBased n)
take m (Cons s k (Shape.ZeroBased n)) =
   if m<0
      then error $ printf "Slice.take: negative size %d" $ toInteger m
      else Cons s k $ Shape.ZeroBased $ min m n

{- |
prop> :{
   QC.forAll (genSlice =<< genAppend) $ \slice ->
   Slice.drop
      (case Slice.shape slice of (sh::+_) -> Shape.size sh)
      (toShapeInt slice)
   ==
   toShapeInt (Slice.right slice)
:}

It is an unchecked error,
when the values of type @n@ exceed the range of @Int@.
-}
drop :: (Slice lay, Integral n) =>
   n -> T lay (Shape.ZeroBased n) -> T lay (Shape.ZeroBased n)
drop m (Cons s k (Shape.ZeroBased n)) =
   if m<0
      then error $ printf "Slice.drop: negative size %d" $ toInteger m
      else Cons (s + elemInc k * fromIntegral m) k $ Shape.ZeroBased $ max 0 $ n-m

{- |
The meaning of the parameters is: @Slice.sub start size@

prop> :{
   \(QC.NonNegative n) (QC.NonNegative start) (QC.NonNegative size) ->
   QC.forAll (genSlice $ shapeInt n) $ \slice ->
   Slice.sub start size slice == Slice.take size (Slice.drop start slice)
:}

It is an unchecked error,
when the values of type @n@ exceed the range of @Int@.
-}
sub :: (Slice lay, Integral n) =>
   n -> n -> T lay (Shape.ZeroBased n) -> T lay (Shape.ZeroBased n)
sub start_ size (Cons s k (Shape.ZeroBased n)) =
   if' (start_<0)
         (error $ printf "View.sub: negative start %d" $ toInteger start_) $
   if' (size  <0)
         (error $ printf "View.sub: negative size %d" $ toInteger size) $
   Cons (s + elemInc k * fromIntegral start_) k $
      Shape.ZeroBased $ max 0 $ min size $ n-start_

{- |
Take a range that must be contained in the input range.

prop> :{
   \n m ->
   let rng = Shape.Range (min n m) (max n (m::Int)) in
   QC.forAll (genSlice rng) $ \slice ->
   Slice.subRange rng slice == slice
:}

It is an unchecked error,
when the values of type @n@ exceed the domain of @Int@.
-}
subRange :: (Slice lay, Ix n) =>
   Shape.Range n -> T lay (Shape.Range n) -> T lay (Shape.Range n)
subRange rng@(Shape.Range from1 to1) (Cons s k (Shape.Range from0 to0)) =
   if' (not $ Ix.inRange (from0,to0) from1)
         (error "View.subRange: new range start is outside old range") $
   if' (not $ Ix.inRange (from0,to0) to1)
         (error "View.subRange: new range end is outside old range") $
   Cons (s + elemInc k * Ix.index (from0,to0) from1) k rng

{- |
Take a range that must be contained in the input range.

prop> :{
   \n m ->
   let rng = Shape.Interval (min n m) (max n (m::Int)) in
   QC.forAll (genSlice rng) $ \slice ->
   Slice.subInterval rng slice == slice
:}

It is an unchecked error,
when the values of type @n@ exceed the domain of @Int@.
-}
subInterval :: (Slice lay, Integral n) =>
   Shape.Interval n -> T lay (Shape.Interval n) -> T lay (Shape.Interval n)
subInterval
   rng@(Shape.Interval from1 to1)
      (Cons s k (Shape.Interval from0 to0)) =
   if' (from1<from0)
         (error $
          printf
            "View.subInterval: new interval start %d less than old one %d"
               (toInteger from1) (toInteger from0)) $
   if' (to1>to0)
         (error $
          printf
            "View.subInterval: new interval end %d greater than old one %d"
               (toInteger to1) (toInteger to0)) $
   Cons (s + elemInc k * fromIntegral (from1-from0)) k rng


{-
ToDo: NestedTuple. - not needed, because no inner shapes
Possible approaches:
  Slice.snd3 . Slice.fst
  ((x,pick,x),x) - using type vodoo to prevent more than one pick
  dissect :: T (NestedTuple (X,X,X)) -> (T,T)
-}

{- |
Convert Map-shaped slice into a dictionary of sub-shapes.
We do not provide a function for picking a single shape,
because that function would have to compute
the sizes of all shapes to the left, anyway.

prop> :{
   QC.forAll (fmap shapeInt $ QC.choose (0,100)) $ \shapeA ->
   QC.forAll (fmap shapeInt $ QC.choose (0,100)) $ \shapeB ->
   QC.forAll (fmap shapeInt $ QC.choose (0,100)) $ \shapeC ->
   QC.forAll (genSlice2
                  (Map.fromList $
                     ('a', shapeA) : ('b', shapeB) : ('c', shapeC) : [])
                  (shapeA ::+ shapeB ::+ shapeC)) $
      \(sliceMap, sliceParted) ->
   Slice.map sliceMap Map.! 'b' == Slice.left (Slice.right sliceParted)
:}

-- FixMe: rowShape and columnShape names are misleading
prop> :{
   QC.forAll (QC.choose (0,100)) $ \numRows ->
   QC.forAll (QC.choose (0,100)) $ \numColumns ->
   let rowShape = shapeInt numRows
       columnShape = shapeInt numColumns
       mapShape_ =
         Map.fromList $ map (\k -> (k, columnShape)) (Shape.indices rowShape) in
   QC.forAll (genSlice2 mapShape_ (rowShape, columnShape)) $
      \(sliceMap, sliceMatrix) ->

   Map.toAscList (Slice.map sliceMap)
   ==
   Array.toAssociations (Slice.rowArray sliceMatrix)
-}
map :: (Slice lay, Shape.C sh) => T lay (Map k sh) -> Map k (T lay sh)
map = traversable

intMap :: (Slice lay, Shape.C sh) => T lay (IntMap sh) -> IntMap (T lay sh)
intMap = traversable

list :: (Slice lay, Shape.C sh) => T lay [sh] -> [T lay sh]
list = traversable

-- ToDo: maybe make private
traversable ::
   (Trav.Traversable f, Slice lay, Shape.C sh) => T lay (f sh) -> f (T lay sh)
traversable (Cons s k m) =
   snd $
   Trav.mapAccumL
      (\offset sh -> (offset + Shape.size sh * elemInc k, Cons offset k sh)) s m
